/*-
 * Copyright (C) @BABOLO  2011 Nov 13 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2011 Nov 13 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: mesraw.c,v 1.4 2014/10/18 09:43:38 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT 3
#define MIFE_COMPAT 4
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "messarge.h"
#include "mesraw.h"

static messarge_raw messarge =
{ BLIN_BIT0
, MESSARGE_MAXH
, MESSARGE_HLEN
, 0
, 0
, 1
, 0
, 0
, NULL
, NULL
, NULL
, 0
, 0
, { { 0, 0,   0, ""}
  , { 0, 0, 'F', "From "}
  , { 0, 0, 'f', "From: "}
  , { 0, 0, 's', "Subject: "}
  , { 0, 0, 't', "To: "}
  , { 0, 0, 'c', "Cc: "}
  , { 0, 0, 'b', "Bcc: "}
  , { 0, 0, 'd', "Date: "}
  , { 0, 0, 'r', "In-Reply-To: "}
  , { 0, 0, 'i', "Message-Id: "}
  , { 0, 0, 'y', "Content-Type: "}
  , { 0, 0, 'e', "Received: "}
} };

static int
doio(messarge_raw *messarge) {
    int ex = 1;

#   define blin_internal_flags (messarge->flags & BLIN_MASK)
    messarge->buflen = mife_ctl(messarge->md, MIFE_CTLREAD, messarge->offset, (ssize_t)0);
    if  (messarge->buflen < 0) {
        ifBLIN_QW0("mife CTLREAD");
        ex = -EX_IOERR;
        goto out;
    }
    if  (messarge->buflen == 0) {
        ex = -EX_OK;
        goto out;
    }
    if  (!(messarge->buffer = mife_point(messarge->md, MIFE_POINOFF, messarge->offset))) {
        ifBLIN_QW0("mife POINOFF");
        ex = -EX_IOERR;
        goto out;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static ssize_t
awrit(mife_aout *ma, const void *buf, ssize_t nbytes) {
    const char *c = buf;
    ssize_t p;

#   define blin_internal_flags (ma->flag & BLIN_MASK)
    for (p = 0; p < nbytes; p++) {
        if  (mife_aput(ma, *(c++)) < 0) {
            ifBLIN_QW0("mife_aput #2");
            p = -1;
            break;
    }   }
    return(p);
#   undef blin_internal_flags
}

static int
gch(messarge_raw *messarge, int *c) {
/*****************************************************************************************************
 *      .  buffer, offset, buflen.                    *
 *   ,   EOF.   messarge->ma    c.                        *
 * c == -2                                                                                  *
 * c == -1                                                                                    *
 * c ==  0  ,   c = -1.                                           *
 * c  >  0  ,  c--                                                             *
 * c  > -2  EOF -  \n.                                                                       *
 *****************************************************************************************************/
    char **p = (char**)&(messarge->buffer);
    int ex = EOF, s;

#   define blin_internal_flags (messarge->flags & BLIN_MASK)
    if  (messarge->buflen) {
        s = ex = *((*p)++) & 0xFF;
        messarge->offset++;
        messarge->buflen--;
    } else {
        ex = EOF;
        s = '\n';
    }
    if  (!*c && (ex != ' ') && (ex != '\t')) --*c;
    if  (*c > 0) --*c;
    if  (((*c == -1) || (ex < 0)) && (mife_aput(messarge->ma, s) < 0)) {
        ifBLIN_QW0("mife_aput #1");
        ex = EOF;
    }
    return(ex);
#   undef blin_internal_flags
}

static int
doespr(messarge_raw *messarge, messarge_headers h) {
/*****************************************************************************************************
 * ,    .                                                           *
 * doespr() == -2                                                                          *
 * doespr() == -1                                                                         *
 * doespr() ==  0  ,  gch()  -1                                       *
 * doespr()  >  0  doespr() ,  gch()  0 -  .           *
 *****************************************************************************************************/
#   define Hb 0x01 /* test body    */
#   define Hn 0x02 /* test any hdr */
#   define Ha 0x04 /* test allhdr  */
#   define Hc 0x08 /* test count   */
#   define Hm 0x10 /* count--      */
#   define Hl 0x20 /* hdr len      */

    u_int32_t at[] =
    { Hb |      Hm      /* h == 0 body           */
    , Ha | Hc | Hm | Hl /* h in known headers    */
    , Ha                /* h == MESSARGE_MAXH    */
    , Hb | Hn           /* h == MESSARGE_EMPTYLN */
    };
    u_int32_t hh;
    int c = -2;

#   define blin_internal_flags (messarge->flags & BLIN_MASK)
    hh = (h < MESSARGE_MAXH) ? !!h : h - MESSARGE_MAXH + 2;
    hh = (hh < 4) ? at[hh] : 0;
    if  ((hh & Hb) && messarge->h_about[0].countr) c = -1;
    if  ((hh & Hb) && !messarge->hdryes) c = -2;
    if  ((hh & Ha) && (messarge->flags & MESSARGE_ALLHDR)) c = -1;
    if  ((hh & Hc) && messarge->h_about[h].countr) c = 0;
    if  ((hh & Hm) && messarge->h_about[h].countr) messarge->h_about[h].countr--;
    if  (messarge->bemail && (messarge->bemail != messarge->domail)) c = -2;
    if  ((hh & Hl) && !c && !(messarge->flags & MESSARGE_EXACT)) {
        c = strlen(messarge->h_about[h].header);
    }
    return(c);
#   undef blin_internal_flags
}

int
messarge_parser(messarge_raw *messarge) {
    messarge_headers h;
    int ex = EX_OK;
    int32_t hc;
    int c;

#   define blin_internal_flags (messarge->flags & BLIN_MASK)
    while ((ex = doio(messarge)) > 0) {
        c = -2;
        if  (babolo_testword(&hdrs, messarge->buffer) != MESSARGE_From) {
            ifBLIN_QX1("Mbox format error at %d line\n", messarge->ln);
            do {
                do {
                    if  ((ex = gch(messarge, &c)) < 0) goto out;
                } while(ex != '\n');
                messarge->ln++;
                if  ((ex = doio(messarge)) <= 0) goto out;
            } while (babolo_testword(&hdrs, messarge->buffer) != MESSARGE_From);
        }
        hc = !!(messarge->flags & MESSARGE_ALLHDR);
        for (h = 0; h < MESSARGE_MAXH; h++) {
            messarge->h_about[h].countr = messarge->h_about[h].track;
            if  (hc < messarge->h_about[h].countr) hc = messarge->h_about[h].countr;
        }
        messarge->domail++;
        /* From */
        h = MESSARGE_From;
        c = doespr(messarge, h);
        ifBLIN_QX4("h%d c%d %"BLIN_X" %"BLIN_O"d", h, c, messarge->buffer, messarge->offset);
        if  (awrit(messarge->ma, messarge->h_about[h].header, c) < 0) {
            ifBLIN_QX0("awrit");
            ex = -EX_IOERR;
            goto out;
        }
        do {
            if  ((ex = gch(messarge, &c)) < 0) goto out;
        } while(ex != '\n');
        messarge->ln++;
        for (;;) { /* headers */
            if  ((ex = doio(messarge)) <= 0) goto out;
            h = babolo_testword(&hdrs, messarge->buffer);
            c = doespr(messarge, h);
            ifBLIN_QX4("h%d c%d %"BLIN_X" %"BLIN_O"d", h, c, messarge->buffer, messarge->offset);
            if  (h == MESSARGE_EMPTYLN) {
                if  ((c) && (mife_aput(messarge->ma, '\n') < 0)) {
                }
                break;
            }
        }
    }
out:
    return(ex);
#   undef blin_internal_flags
}

void
usage(messarge_raw *messarge, int ex) {
    messarge_headers h;
    int i = 0;

#   define blin_internal_flags (messarge->flags & BLIN_MASK)
    ifBLIN_QO0 i++;
    if  (i || !ex) {
        fprintf( stderr
               , "messarge @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
                 "Usage:\n"
                 "%s -<options> [num]\n"
                 "     num -       .\n"
                 "     -C -      ,\n"
                 "             .\n"
                 "     -= -    .\n"
                 "     -h -   .\n"
                 "     -n -         .\n"
                 "             1.\n"
                 "     -q -       ,   \n"
                 "             ,     \n"
                 "     -Q -      stderr\n"
                 "     -v -   \n"
                 "     -z -     \n"
                 "     -Z -     \n"
               , getprogname()
           );
        for (h = MESSARGE_From; h < MESSARGE_MAXH; h++) {
            fprintf( stderr, "     -%c -   %s\n"
                   , messarge->h_about[h].flag, messarge->h_about[h].header
                   );
    }   }
    if  (ex) exit(ex);
#   undef blin_internal_flags
}

int
main(int argc, char **argv) {
    ifBLIN_FLAG = messarge.flags;
    babolo_opts *bos = NULL;
    messarge_headers h;
    int c, ex = EX_OK;
    u_int32_t countr;               /*     */
    const char *o;

    if  (!(bos = babolo_openopts(BLIN_BIT0 | Bpars_SYME | Bpars_ESYM, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        usage(&messarge, EX_SOFTWARE);
    }
    if  (babolo_setopts(bos, BLIN_BIT0 | Bpars_NONU, argc, argv, "=bcC:deFfhinqQrstvy?")) {
        ifBLIN_QW0("babolo_setopts");
        usage(&messarge, EX_SOFTWARE);
    }
    while (!!(c = babolo_getopts(bos))) {
        switch (c) {
        case 'C': countr = babolo_getoptunum(bos, 0);  break;
        case '=': messarge.flags |= MESSARGE_EXACT;    break;
        case 'h': messarge.flags |= MESSARGE_ALLHDR;   break;
        case 'n': messarge.flags |= MESSARGE_NUMERATE; break;
        case 'q': messarge.flags |= MESSARGE_QUIET;    break;
        case 'Q': BLIN_QUIET(messarge.flags);          break;
        case 'v': BLIN_VERBOSE(messarge.flags);        break;
        case '?': usage(&messarge, EX_OK);
        default : for (h = MESSARGE_From; h < MESSARGE_MAXH; h++) {
                      if  (messarge.h_about[h].flag == c) {
                          messarge.h_about[h].track++;
                          if  (messarge.hdryes < messarge.h_about[h].track)
                              messarge.hdryes = messarge.h_about[h].track;
                          break;
                  }   }
                  if  (h == MESSARGE_MAXH) {
                      ifBLIN_QX0("Unknown option -%c", c);
                      usage(&messarge, EX_USAGE);
    }   }         }
    messarge.bemail = messarge.domail = 0;
    errno = 0;
    if  ((o = babolo_getargs(bos))) {
        messarge.bemail = atoi(o);
    } else if (errno) {
        ifBLIN_QW0("babolo_getargs #1");
        ex = EX_SOFTWARE;
        goto out;
    }
    errno = 0;
    if  ((o = babolo_getargs(bos))) {
        ifBLIN_QX0("Only 1 parameter accepted, more provided");
        ex = EX_USAGE;
        goto out;
    } else if (errno) {
        ifBLIN_QW0("babolo_getarg #2");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!messarge.hdryes && (messarge.flags & MESSARGE_ALLHDR)) messarge.hdryes = 1;
    if  (!((messarge.flags & MESSARGE_QUIET) && messarge.hdryes)) {
        messarge.h_about[0].track = -1;
    }   /*   -q          */
    if  (!(messarge.md = mife_init(MIFE_SLID))) {
        ifBLIN_QW0("mife_init");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctl(messarge.md, MIFE_CTLFDSC, fileno(stdin))) {
        ifBLIN_QW0("mife CTLFDSC");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctl(messarge.md, MIFE_CTLEOLI, "\n")) {
        ifBLIN_QW0("mife CTLEOLI");
        ex = EX_IOERR;
        goto out;
    }
    messarge.offset = 0;
    messarge.ma = mife_aopen(kukMIFE_BUFL(3) | MIFE_CLOS, fileno(stdout));
    if  (!messarge.ma) {
        ifBLIN_QW0("mife_aopen");
        ex = EX_IOERR;
        goto out1;
    }
    ex = messarge_parser(&messarge);
    if  (ex < -1) ex = -ex; else ex = 0;
    if  (mife_aclose(messarge.ma) < 0) {
        ifBLIN_QW0("mife_aclose");
        ex = EX_IOERR;
        goto out;
    }
out1:
    if  (mife_ctl(messarge.md, MIFE_CTLFINI) < 0) {
        ifBLIN_QW0("mife CTLFINI");
        ex = EX_IOERR;
        goto out;
    }
out:
    exit(ex);
}
