/*-
 * Copyright (C)1999..2011 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: messarge.h,v 1.12 2014/10/17 08:37:38 babolo Exp $
 */

#include <sys/errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sysexits.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <fcntl.h>
#include <regex.h>
#include <stdio.h>
#include <time.h>

#ifndef STRTOQ
#   define STRTOQ strtoq
#endif
#ifndef P_MASK
#   define P_MASK 017777  /* page size for mmap. be careful */
#endif

#define HDR_FLAG 000001
#define WWW_FLAG 000002
#define BIN_FLAG 000004

#define WORK_SIZE 1024
#define STATE_NAME_LEN 32
#define DB_VAR_MAX_SIZE 1024
#define DB_PATH_SIZE 1024
#define DB_VAR_SIZE 183
#define MAX_L_NUMBER 32
#define MASK_T unsigned            /* Depends on MAX_L_NUMBER */
#define DB_BLOCK_SIZE ((void*)db->entry.var - (void*)&db->entry + db->header.var_size)
#define RE_EXCL_RE "^((RE:|RE\\[[0-9]*\\]:|Subject:|[H][A]:|FW:|:|OTP:|:)[ \t]*)*"
#define RE_STRING "> "
#define ENTRY(a) ((struct db_entry*)\
                     (       ((void*)entries)\
                     +   (   (   (void*)db->entry.var\
                             -   (void*)&db->entry\
                             +   ((struct db_header*)entries)->var_size\
                             )\
                         *   (((struct db_header*)entries)->len + (a) - 1)\
                 )   )   )
#define VAR_STR (&(entry->var[db->header.LList]))
#define MAGIK_00 0x0062446d
#define MAGIK_01 0x0162446d
#define MAGIK_VV MAGIK_01

#define MESSARGE_QUIET         0x100000
#define MESSARGE_NUMERATE      0x200000
#define MESSARGE_ALLHDR        0x400000
#define MESSARGE_EXACT         0x800000
#define MESSARGE_HLEN          31

typedef enum messarge_headers
{ MESSARGE_From         = 1
, MESSARGE_from         = 2
, MESSARGE_subject      = 3
, MESSARGE_to           = 4
, MESSARGE_cc           = 5
, MESSARGE_bcc          = 6
, MESSARGE_date         = 7
, MESSARGE_in_reply_to  = 8
, MESSARGE_message_id   = 9
, MESSARGE_content_type = 10
, MESSARGE_received     = 11
, MESSARGE_MAXH         = 12
, MESSARGE_EMPTYLN      = 13
} messarge_headers;

#ifdef _MIFE_H_
#ifdef _BABOLO_BLINflag_H_
typedef struct messarge_raw {
    BLIN_flag        flags;
    u_int32_t        maxh;     /*    */
    u_int32_t        hlen;     /*        */
    u_int32_t        bemail;   /*               */
    u_int32_t        domail;   /*                   */
    u_int32_t        ln;       /*              */
    off_t            offset;   /*          */
    ssize_t          buflen;   /*       */
    mife_descriptor *md;       /*                           */
    void            *buffer;   /*                 */
    mife_aout       *ma;       /*                          */
    u_int32_t        hdryes;   /*  countr                  */
    u_int32_t        dummy;
    struct {
        int32_t  countr;
        int32_t  track;
        u_char   flag;
        char     header[MESSARGE_HLEN];
    }                h_about[MESSARGE_MAXH];
} messarge_raw;
#endif
#endif

struct db_header_00 {
    unsigned magic;
    unsigned len;
    unsigned N_start;
    unsigned var_size;
    unsigned flags;
#       define PATH_IS_PIPE     1
#       define MAKE_ON_WAKEUP   2
    unsigned char NLists, LList;
    char path[DB_PATH_SIZE];
};

struct db_header {
    unsigned magic;
    unsigned len;
    unsigned N_start;
    unsigned var_size;
    unsigned flags;
#       define PATH_IS_PIPE     1
#       define MAKE_ON_WAKEUP   2
    unsigned char NLists, LList;
    char path[DB_PATH_SIZE];
};

struct db_entry_00 {
    unsigned long long subj_hash, mail_hash, repl_hash, id_hash;
    unsigned      offset  /* this message's offset in mailbox            */
    ,            headerl  /* this message's header length                */
    ,              bodyl  /* this message's full length                  */
    ;
    time_t     mail_time; /* time from 'Data: ' line                     */
    time_t   stored_time; /* time entry created                          */
    unsigned        plex  /* entry for previous sorted by subject entry  */
    ,               lext  /* entry for next sorted by subject entry      */
    ;
    short           prev  /* entry offset for previous same level entry  */
    ,               next  /* entry offset for next same level entry      */
    ;
    char           right; /* nest level offset in prev/next chain        */
    unsigned char     th  /* node type in thread                         */
    ,            charset  /* charset in Content-Type: text               */
    ,      encapsulation  /* Content-Transfer-Encoding            */
    ,var[DB_VAR_MAX_SIZE] /* bit mask of lists this entry belongs */
    ;                     /* sender's address                     */
                          /* sender's name                        */
                          /* subject                              */
};

struct db_entry {
    unsigned long long subj_hash, mail_hash, repl_hash, id_hash;
    unsigned      offset  /* this message's offset in mailbox            */
    ,            headerl  /* this message's header length                */
    ,              bodyl  /* this message's full length                  */
    ,              bound  /**/
    ;
    time_t     mail_time; /* time from 'Data: ' line                     */
    time_t   stored_time; /* time entry created                          */
    unsigned        plex  /* entry for previous sorted by subject entry  */
    ,               lext  /* entry for next sorted by subject entry      */
    ;
    short           prev  /* entry offset for previous same level entry  */
    ,               next  /* entry offset for next same level entry      */
    ;
    unsigned char boundl; /**/
    char           right; /* nest level offset in prev/next chain        */
    unsigned char     th  /* node type in thread                         */
    ,            charset  /* charset in Content-Type: text               */
#        define CODEMASK 0200  /* Content-Type: text/html                */
    ,      encapsulation  /* Content-Transfer-Encoding            */
    ,var[DB_VAR_MAX_SIZE];/* bit mask of lists this entry belongs */
                          /* sender's address                     */
                          /* sender's name                        */
                          /* subject                              */
};

struct db {
    struct db_header header;
    struct db_entry entry;
};

struct shformat {
    unsigned char *L_str;
    unsigned char *R_str;
    unsigned char *T_str;
    unsigned char *B_str;
    unsigned char *E_str;
    unsigned char *A_str;
};

struct tinterval {
    time_t tmin;
    time_t tmax;
    int nomin;
    int nomax;
};

enum th_state {th_root, th_cousen, th_brother, th_child, th_hidden, th_purge} th_state;

enum code
{ no_code
, code_k
, code_a
, code_w
, code_i
  /* place for new charsets */
, code_unknown=127
} code;

enum encapsulation {none_enc, bit7, bit8, binary, quoted, base64, uu};
enum quost {q_start, q_eq, q_1, q_xxx};
enum qpro {qp_skip, qp_thru, qp_stor, qp_hex, qp_sp, qp_err, qp_ers};
void exr(regex_t *preg, const char *pattern, int cflags);
void perr(int eval, const char *fmt);
