/*-
 * Copyright (C)1999..2011 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)1999..2011 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: mestun.c,v 1.6 2014/10/10 11:47:10 babolo Exp $";
#endif /* not lint */

#include "messarge.h"
#include "mescommon.h"

enum op {op_none, op_timet, op_mhash, op_mchg, op_list, op_purge} op_do;
extern int v_flag;

int
can_chain(unsigned c, unsigned i) {
    if  (c == 0 || i == 0) return(1);
    if  (c > i) return((c - i) <= SHRT_MAX); else  return((i - c) <= SHRT_MAX);
}

void
usage(int exitcode) {
    fprintf( stderr
           , "messarge @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "%s -[v]t <window size> <file IDX>\n"
             "         for adjust store times if start with non empty mailbox\n"
             "%s -m[v] <file IDX> <file IDX with new hashes and boundaries>\n"
             "         for repare INDX created by old messarge versions\n"
             "%s -c[v] <file IDX> <file IDX with new indexint times>\n"
             "         for repare INDX created by old messarge versions\n"
             "%s -[v]l <list number> <file IDX> [<messageN...>]\n"
             "         mark some messages with <list number> or clear any such mark\n"
             "         when <list number> == 0\n"
             "%s -p[v] <file IDX> [<messageN...>]\n"
             "         purge given messages from index\n"
           , getprogname(), getprogname(), getprogname(), getprogname(), getprogname()
           )
     ;
     exit(exitcode);
}

int
main(int argc, char **argv) {
    int i, j, c, win_t = 0, listN = 0;
    unsigned k, n;
    struct stat sb;
    struct timeval tv[2];
    char *idxname;
    off_t a;
#ifdef DEBUG
# if DEBUG > 1
    char argv0[] = "mestun";
    char argv1[] = "-m";
    char argv2[] = "tmpm";
    char argv3[] = "tmpn";
    char *arg[] = {argv0, argv1, argv2, argv3, NULL};
    argv = arg;
    argc = 4;
# endif
#endif

    opterr = 0;
    op_do = op_none;
    while ((c = getopt(argc, argv, "vt:mcl:p?")) !=  -1) {
        switch (c) {
        case 'v':
            v_flag++;
            break;
        case 't':
            op_do = op_timet;
            win_t = (int)strtol(optarg, NULL, 0);
            break;
        case 'm':
            op_do = op_mhash;
            break;
        case 'c':
            op_do = op_mchg;
            break;
        case 'l':
            op_do = op_list;
            listN = (int)strtol(optarg, NULL, 0);
            break;
        case 'p':
            op_do = op_purge;
            break;
        default: usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (op_do == op_none) usage(EX_USAGE);
    db = calloc(1, sizeof(struct db));
    if  (stat(argv[0], &sb))                                                     perr(EX_NOINPUT, NULL);
    idxname = argv[0];
    tv[0].tv_sec = sb.st_atime;
    tv[1].tv_sec = sb.st_mtime;
#ifdef NO_NSEC
    tv[0].tv_usec = 0;
    tv[1].tv_usec = 0;
#else
    tv[0].tv_usec = sb.st_atimespec.tv_nsec / 1000;
    tv[1].tv_usec = sb.st_mtimespec.tv_nsec / 1000;
#endif
    arcsize = open_db(argv[0], update_m);
    argc -= 1;
    argv += 1;
    arcsize = arcsize / DB_BLOCK_SIZE - db->header.len;
    if  ((void *)( entries
                 = (struct db_entry*)mmap( NULL
                                         , (arcsize + db->header.len) * DB_BLOCK_SIZE
                                         , PROT_READ | PROT_WRITE, MAP_SHARED, filebdes, 0
                                         )
                 ) == MAP_FAILED
        )                                                                        perr(EX_NOINPUT, NULL);
    if  (close(filebdes))                                                        perr(EX_IOERR, NULL);
    switch (op_do) {
    case op_timet:
        if  (argc != 0) usage(EX_USAGE);
        for (i = 1; i <= arcsize; i++) {
            if  (win_t) {
                time_t *tim;
                int min, max;
                unsigned long long sum;

                tim = malloc(sizeof(time_t) * (2 * win_t + 1));
                for (j = -win_t; j <= win_t; j++) {
                    tim[j + win_t] = ((i + j > 0) && (i + j <= arcsize)) ? ENTRY(i + j)->mail_time : 0;
                }
                min = max = 0;
                for (j = -win_t; j <= win_t; j++) {
                    if  (tim[j + win_t] && tim[max + win_t] < ENTRY(i + j)->mail_time) max = j;
                    if  (tim[j + win_t] && tim[min + win_t] > ENTRY(i + j)->mail_time) min = j;
                }
                tim[max + win_t] = tim[min + win_t] = 0;
                sum = 0; k = 0;
                for (j = -win_t; j <= win_t; j++) {
                    if  (tim[j + win_t]) {
                        sum += tim[j + win_t];
                        k++;
                }   }
                ENTRY(i)->stored_time = k ? sum / k : 0;
            } else {
                ENTRY(i)->stored_time = ENTRY(i)->mail_time;
        }   }
        break;
    case op_mhash:
        if  (argc != 1) usage(EX_USAGE);
        a = open_db(argv[0], read_m);
        a = a / DB_BLOCK_SIZE - db->header.len;
        if  (a < arcsize) arcsize = a;
        for (n = 1; n <= arcsize; n++) {
            if  (v_flag) fprintf(stderr, "-m %u\n", n);
            if  (1 != fread(&db->entry, blocksize, 1, DB_F))                      perr(EX_IOERR, NULL);
            if  (  db->entry.offset != ENTRY(n)->offset
                || db->entry.bodyl != ENTRY(n)->bodyl
                ) {
                fprintf(stderr, "%s: index from another mailbox is not applicable\n", getprogname());
                exit(EX_DATAERR);
            }
            ENTRY(n)->subj_hash = db->entry.subj_hash;
            ENTRY(n)->mail_hash = db->entry.mail_hash;
            ENTRY(n)->repl_hash = db->entry.repl_hash;
            ENTRY(n)->id_hash = db->entry.id_hash;
            ENTRY(n)->bound = db->entry.bound;
            ENTRY(n)->boundl = db->entry.boundl;
            ENTRY(n)->charset = db->entry.charset;
            ENTRY(n)->encapsulation = db->entry.encapsulation;
            ENTRY(n)->headerl = db->entry.headerl;
        }
        break;
    case op_mchg:
        if  (argc != 1) usage(EX_USAGE);
        a = open_db(argv[0], read_m);
        a = a / DB_BLOCK_SIZE - db->header.len;
        if  (a < arcsize) arcsize = a;
        for (n = 1; n <= arcsize; n++) {
            if  (v_flag) fprintf(stderr, "-m %u\n", n);
            if  (1 != fread(&db->entry, blocksize, 1, DB_F))                      perr(EX_IOERR, NULL);
            if  (  db->entry.offset != ENTRY(n)->offset
                || db->entry.bodyl != ENTRY(n)->bodyl
                ) {
                fprintf(stderr, "%s: index from another mailbox is not applicable\n", getprogname());
                exit(EX_DATAERR);
            }
            ENTRY(n)->stored_time = db->entry.stored_time;
        }
        break;
    case op_list:
        if  (listN > db->header.NLists) {
            fprintf(stderr, "%s: used list N %u when max %u\n", getprogname(), (unsigned)listN, (unsigned)(db->header.NLists));
            exit(EX_DATAERR);
        }
        j = listN / 8;
        i = listN % 8;
        for (k = 1; i > 1; i--) k *= 2;
        if  (argc == 0) {
            for (n = 1; n <= arcsize; n++) {
                if  (listN) {
                    ENTRY(n)->var[j] |= k;
                } else {
                    for (j = 0; j < db->header.LList; j++) ENTRY(n)->var[j] = 0;
            }   }
        } else {
            while (argc-- > 0) {
                n = (unsigned)strtol(*argv++, NULL, 0);
                if  (v_flag) fprintf(stderr, "N%u ListM=%X\n", n, listN ? (k << (8 * j)) : 0);
                if  (n > db->header.N_start && n <= db->header.N_start + arcsize) {
                    if  (listN) {
                        ENTRY(n - db->header.N_start)->var[j] |= k;
                    } else {
                        for (j = 0; j < db->header.LList; j++)
                            ENTRY(n - db->header.N_start)->var[j] = 0;
                    }
                } else {
                    fprintf( stderr, "messageN must be from %u to %u\n"
                           , db->header.N_start + 1
                           , db->header.N_start + (unsigned)arcsize
                           );
        }   }   }
        break;
    case op_purge:
            while (argc-- > 0) {
                n = (unsigned)strtol(*argv++, NULL, 0);
                if  (v_flag) fprintf(stderr, "N%u purge\n", n);
                if  (n > db->header.N_start && n <= db->header.N_start + arcsize) {
                    n -= db->header.N_start;
                    if  (  ENTRY(n)->th != th_purge
                        && ENTRY(n)->prev != ENTRY(n)->next
                        ) {
                        unsigned p = n + ENTRY(n)->prev;
                        unsigned f = n + ENTRY(n)->next;

                        if  (p + ENTRY(p)->next == n) {
                            if  (can_chain(p, f)) ENTRY(p)->next = f - p; else ENTRY(p)->next = 0;
                        }
                        if  (f + ENTRY(f)->prev == n) {
                            if  (can_chain(p, f)) ENTRY(f)->prev = p - f; else ENTRY(f)->prev = 0;
                        }
                        ENTRY(f)->right += ENTRY(n)->right;
                    }
                    ENTRY(n)->th = th_purge;
                    ENTRY(n)->right = 0;
                    ENTRY(n)->prev = 0;
                    ENTRY(n)->next = 0;
                    ENTRY(n)->plex = 0;
                    ENTRY(n)->lext = 0;
                } else {
                    fprintf( stderr, "messageN must be from %u to %u\n"
                           , db->header.N_start + 1
                           , db->header.N_start + (unsigned)arcsize
                           );
            }   }
        break;
    default:
        fprintf(stderr, "%s: Not yet\n", getprogname());
        exit(EX_OK);
    }
    if  (munmap((void*)entries, (maxmail + db->header.len) * DB_BLOCK_SIZE))     perr(EX_IOERR, NULL);
    if  (utimes(idxname, tv))                                                    perr(EX_IOERR, NULL);
    exit(EX_OK);
}
